import numpy as np

# Planck constants in J s
PLANCK_CONSTANT = 6.62607015e-34
REDUCED_PLANCK_CONSTANT = 1.054571817e-34
# Boltzmann constant in J/K
BOLTZMANN_CONSTANT = 1.380649e-23
# Speed of light in m/s
SPEED_OF_LIGHT = 2.99792458e8
# Electron charge in C
ELECTRON_CHARGE = 1.602176634e-19
# Permeability of free space in N/A^2
PERMEABILITY_OF_FREE_SPACE = 4 * np.pi * 1e-7
# Permittivity of free space in F/m
PERMITTIVITY_OF_FREE_SPACE = 8.854187817e-12
